import timeit

import numpy as np
from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    KARABO_CLASSINFO,
    STRING_ELEMENT,
    TABLE_ELEMENT,
    UINT32_ELEMENT,
    Hash,
    Schema,
)
from karabo.common.states import State

from . import utils
from ._version import version as deviceVersion
from .agipd_gpu import AgipdGainMode, AgipdGpuRunner, BadPixelValues, CorrectionFlags
from .base_correction import BaseCorrection


@KARABO_CLASSINFO("AgipdCorrection", deviceVersion)
class AgipdCorrection(BaseCorrection):
    # subclass *must* set these attributes
    _correction_flag_class = CorrectionFlags
    _correction_slot_names = (
        ("thresholding", CorrectionFlags.THRESHOLD),
        ("offset", CorrectionFlags.OFFSET),
        ("relGainPc", CorrectionFlags.REL_GAIN_PC),
        ("relGainXray", CorrectionFlags.REL_GAIN_XRAY),
        ("badpixels", CorrectionFlags.BPMASK),
    )
    _gpu_runner_class = AgipdGpuRunner

    @staticmethod
    def expectedParameters(expected):
        AgipdCorrection.addConstant(
            "ThresholdsDark",
            "Dark",
            expected,
            optional=False,
            mandatoryForIteration=True,
        )
        AgipdCorrection.addConstant(
            "Offset", "Dark", expected, optional=False, mandatoryForIteration=True
        )
        AgipdCorrection.addConstant(
            "SlopesPC", "Dark", expected, optional=True, mandatoryForIteration=True
        )
        AgipdCorrection.addConstant(
            "SlopesFF",
            "Illuminated",
            expected,
            optional=True,
            mandatoryForIteration=True,
        )
        AgipdCorrection.addConstant(
            "BadPixelsDark", "Dark", expected, optional=True, mandatoryForIteration=True
        )
        AgipdCorrection.addConstant(
            "BadPixelsPC", "Dark", expected, optional=True, mandatoryForIteration=True
        )
        AgipdCorrection.addConstant(
            "BadPixelsFF",
            "Illuminated",
            expected,
            optional=True,
            mandatoryForIteration=True,
        )
        super(AgipdCorrection, AgipdCorrection).expectedParameters(expected)
        for slot_name, _ in AgipdCorrection._correction_slot_names:
            (
                BOOL_ELEMENT(expected)
                .key(f"corrections.available.{slot_name}")
                .readOnly()
                .initialValue(False)
                .commit(),
                BOOL_ELEMENT(expected)
                .key(f"corrections.enabled.{slot_name}")
                .assignmentOptional()
                .defaultValue(False)
                .reconfigurable()
                .commit(),
                BOOL_ELEMENT(expected)
                .key(f"corrections.preview.{slot_name}")
                .assignmentOptional()
                .defaultValue(False)
                .reconfigurable()
                .commit(),
            )

        (
            BOOL_ELEMENT(expected)
            .key("corrections.overrideMdAdditionalOffset")
            .displayedName("Override md_additional_offset")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),
            FLOAT_ELEMENT(expected)
            .key("corrections.mdAdditionalOffset")
            .displayedName("Value for md_additional_offset (if overriding)")
            .description(
                "Normally, md_additional_offset (part of relative gain correction) is "
                "computed when loading SlopesPC. In case you want to use a different "
                "value (global for all medium gain pixels), you can specify it here "
                "and set corrections.overrideMdAdditionalOffset to True."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
        )
        (
            STRING_ELEMENT(expected)
            .key("gainMode")
            .displayedName("Gain mode")
            .assignmentOptional()
            .defaultValue("ADAPTIVE_GAIN")
            .options("ADAPTIVE_GAIN,FIXED_HIGH_GAIN,FIXED_MEDIUM_GAIN,FIXED_LOW_GAIN")
            .commit()
        )
        # TODO: hook this up to actual correction done
        bad_pixel_selection_schema = Schema()
        (
            STRING_ELEMENT(bad_pixel_selection_schema).key("name").readOnly().commit(),
            UINT32_ELEMENT(bad_pixel_selection_schema).key("value").readOnly().commit(),
            BOOL_ELEMENT(bad_pixel_selection_schema)
            .key("apply")
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),
            TABLE_ELEMENT(expected)
            .key("corrections.badPixelFlagsApplied")
            .setNodeSchema(bad_pixel_selection_schema)
            .assignmentOptional()
            .defaultValue(
                [
                    Hash("name", field.name, "value", field.value, "apply", True)
                    for field in BadPixelValues
                ]
            )
            .reconfigurable()
            .commit(),
        )

    @property
    def input_data_shape(self):
        return (
            self.get("dataFormat.memoryCells"),
            2,
            self.get("dataFormat.pixelsX"),
            self.get("dataFormat.pixelsY"),
        )

    @property
    def output_data_shape(self):
        return utils.shape_after_transpose(
            (
                self.get("dataFormat.memoryCells"),
                self.get("dataFormat.pixelsX"),
                self.get("dataFormat.pixelsY"),
            ),
            self._output_transpose,
        )

    def __init__(self, config):
        # TODO: different gpu runner for fixed gain mode
        self.gain_mode = AgipdGainMode[config.get("gainMode")]
        self._gpu_runner_init_args = {"gain_mode": self.gain_mode}

        super().__init__(config)
        output_axis_order = config.get("dataFormat.outputAxisOrder")
        if output_axis_order == "pixels-fast":
            self._output_transpose = None
        elif output_axis_order == "memorycells-fast":
            self._output_transpose = (2, 1, 0)
        else:
            self._output_transpose = None
        self._update_shapes()
        if config.get("corrections.overrideMdAdditionalOffset"):
            self._override_md_additional_offset = config.get(
                "corrections.mdAdditionalOffset"
            )
        else:
            self._override_md_additional_offset = None
        self.updateState(State.ON)

    def process_input(self, data, metadata):
        """Registered for dataInput, handles all processing and sending"""

        if not self._schema_cache["doAnything"]:
            if self._schema_cache["state"] is State.PROCESSING:
                self.updateState(State.ACTIVE)
            return

        source = metadata.get("source")

        if source not in self.sources:
            self.log.INFO(f"Ignoring unknown source {source}")
            return

        # TODO: what are these empty things for?
        if not data.has("image"):
            # self.log.INFO("Ignoring hash without image node")
            return

        time_start = timeit.default_timer()

        train_id = metadata.getAttribute("timestamp", "tid")
        cell_table = np.squeeze(data.get("image.cellId"))
        assert isinstance(cell_table, np.ndarray), "image.cellId should be ndarray"
        if len(cell_table.shape) == 0:
            msg = "cellId had 0 dimensions. DAQ may not be sending data."
            self.set("status", msg)
            self.log.WARN(msg)
            return
        # original shape: memory_cell, data/raw_gain, x, y
        image_data = data.get("image.data")
        if image_data.shape[0] != self._schema_cache["dataFormat.memoryCells"]:
            self.set(
                "status", f"Updating input shapes based on received {image_data.shape}"
            )
            self.set("dataFormat.memoryCells", image_data.shape[0])
            with self._buffer_lock:
                # TODO: pulse filter update after reimplementation
                self._update_shapes()

        if not self._schema_cache["state"] is State.PROCESSING:
            self.updateState(State.PROCESSING)
            self.set("status", "Processing data")
        if self._state_reset_timer is None:
            self._state_reset_timer = utils.DelayableTimer(
                timeout=self._schema_cache["processingStateTimeout"],
                callback=self._reset_state_from_processing,
            )
        else:
            self._state_reset_timer.set_timeout(
                self._schema_cache["processingStateTimeout"]
            )

        correction_cell_num = self._schema_cache["dataFormat.constantMemoryCells"]
        do_generate_preview = (
            train_id % self._schema_cache["preview.trainIdModulo"] == 0
            and self._schema_cache["preview.enable"]
        )

        with self._buffer_lock:
            # cell_table = cell_table[self.pulse_filter]
            pulse_table = np.squeeze(data.get("image.pulseId"))  # [self.pulse_filter]
            cell_table_max = np.max(cell_table)
            # TODO: all this checking and warning can go into GPU runner
            if cell_table_max >= correction_cell_num:
                msg = (
                    f"Max cell ID ({cell_table_max}) exceeds range for loaded "
                    f"constants ({correction_cell_num} cells). Some frames will not be "
                    "corrected."
                )
                self.log.WARN(msg)
                self.set("status", msg)

            self.gpu_runner.load_data(image_data)
            buffer_handle, buffer_array = self._shmem_buffer.next_slot()
            self.gpu_runner.load_cell_table(cell_table)
            self.gpu_runner.correct(self._correction_flag_enabled)
            self.gpu_runner.reshape(out=buffer_array)
            # after reshape, data for dataOutput is now safe in its own buffer
            if do_generate_preview:
                if self._correction_flag_enabled != self._correction_flag_preview:
                    self.gpu_runner.correct(self._correction_flag_preview)
                preview_slice_index = self._schema_cache["preview.pulse"]
                if preview_slice_index >= 0:
                    # look at pulse_table to find which index this pulse ID is in
                    # TODO: move this to GPU
                    pulse_id_found = np.where(pulse_table == preview_slice_index)[0]
                    if len(pulse_id_found) == 0:
                        pulse_found_instead = pulse_table[0]
                        msg = (
                            f"Pulse {preview_slice_index} not found in "
                            f"image.pulseId, arbitrary pulse "
                            f"{pulse_found_instead} will be shown."
                        )
                        self.log.WARN(msg)
                        self.set("status", msg)
                        preview_slice_index = 0
                    else:
                        preview_slice_index = pulse_id_found[0]
                preview_raw, preview_corrected = self.gpu_runner.compute_preview(
                    preview_slice_index
                )

        data.set("image.data", buffer_handle)
        data.set("image.cellId", cell_table[:, np.newaxis])
        data.set("image.pulseId", pulse_table[:, np.newaxis])
        data.set("calngShmemPaths", ["image.data"])
        self._write_output(data, metadata)
        if do_generate_preview:
            self._write_combiner_preview(
                preview_raw, preview_corrected, train_id, source
            )

        # update rate etc.
        self._buffered_status_update.set("trainId", train_id)
        self._rate_tracker.update()
        time_spent = timeit.default_timer() - time_start
        self._buffered_status_update.set(
            "performance.lastProcessingDuration", time_spent * 1000
        )

    def _load_constant_to_gpu(self, constant_name, constant_data):
        if constant_name == "ThresholdsDark":
            if self.gain_mode is not AgipdGainMode.ADAPTIVE_GAIN:
                self.log.INFO("Loaded ThresholdsDark ignored due to fixed gain mode")
                return
            self.gpu_runner.load_thresholds(constant_data)
            # TODO: encode correction / constant dependencies in a clever way
            if not self.get("corrections.available.thresholding"):
                self.set("corrections.available.thresholding", True)
                self.set("corrections.enabled.thresholding", True)
                self.set("corrections.preview.thresholding", True)
        elif constant_name == "Offset":
            self.gpu_runner.load_offset_map(constant_data)
            if not self.get("corrections.available.offset"):
                self.set("corrections.available.offset", True)
                self.set("corrections.enabled.offset", True)
                self.set("corrections.preview.offset", True)
        elif constant_name == "SlopesPC":
            self.gpu_runner.load_rel_gain_pc_map(constant_data)
            if not self.get("corrections.available.relGainPc"):
                self.set("corrections.available.relGainPc", True)
                self.set("corrections.enabled.relGainPc", True)
                self.set("corrections.preview.relGainPc", True)
            if self._override_md_additional_offset is not None:
                self.gpu_runner.md_additional_offset_gpu.fill(
                    self._override_md_additional_offset
                )
        elif constant_name == "SlopesFF":
            self.gpu_runner.load_rel_gain_ff_map(constant_data)
            if not self.get("corrections.available.relGainXray"):
                self.set("corrections.available.relGainXray", True)
                self.set("corrections.enabled.relGainXray", True)
                self.set("corrections.preview.relGainXray", True)
        elif "BadPixels" in constant_name:
            # TODO: implement loading bad pixels
            self.gpu_runner.load_badpixels_map(constant_data)
            if not self.get("corrections.available.badpixels"):
                self.set("corrections.available.badpixels", True)
                self.set("corrections.enabled.badpixels", True)
                self.set("corrections.preview.badpixels", True)

    def _update_pulse_filter(self, filter_string):
        """Called whenever the pulse filter changes, typically followed by
        _update_shapes"""

        if filter_string.strip() == "":
            new_filter = np.arange(self.get("dataFormat.memoryCells"), dtype=np.uint16)
        else:
            new_filter = np.array(eval(filter_string), dtype=np.uint16)
        assert np.max(new_filter) < self.get("dataFormat.memoryCells")
        self.pulse_filter = new_filter

    def postReconfigure(self):
        super().postReconfigure()
        if self.get("corrections.overrideMdAdditionalOffset"):
            self._override_md_additional_offset = self.get("corrections.mdAdditionalOffset")
            self.gpu_runner.md_additional_offset_gpu.fill(
                self._override_md_additional_offset
            )
        else:
            self._override_md_additional_offset = None
