import enum

import cupy
import numpy as np

from . import base_gpu, utils


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    THRESHOLD = 1
    OFFSET = 2
    BLSHIFT = 4
    REL_GAIN_PC = 8
    REL_GAIN_XRAY = 16
    BPMASK = 32


# from pycalibration's enum.py
class AgipdGainMode(enum.IntEnum):
    ADAPTIVE_GAIN = 0
    FIXED_HIGH_GAIN = 1
    FIXED_MEDIUM_GAIN = 2
    FIXED_LOW_GAIN = 3


class AgipdGpuRunner(base_gpu.BaseGpuRunner):
    _kernel_source_filename = "agipd_gpu_kernels.cpp"

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells,
        output_transpose=(1, 2, 0),  # default: memorycells-fast
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
        badpixel_mask_value=np.float32(np.nan),
        gain_mode=AgipdGainMode.ADAPTIVE_GAIN,
    ):
        self.gain_mode = gain_mode
        if self.gain_mode is AgipdGainMode.ADAPTIVE_GAIN:
            self.default_gain = np.uint8(gain_mode)
        else:
            self.default_gain = np.uint8(gain_mode - 1)
        self.input_shape = (memory_cells, 2, pixels_x, pixels_y)
        self.processed_shape = (memory_cells, pixels_x, pixels_y)
        super().__init__(
            pixels_x,
            pixels_y,
            memory_cells,
            constant_memory_cells,
            output_transpose,
            input_data_dtype,
            output_data_dtype,
        )
        self.gain_map_gpu = cupy.empty(self.processed_shape, dtype=np.uint8)

        self.map_shape = (self.constant_memory_cells, self.pixels_x, self.pixels_y)
        self.gm_map_shape = self.map_shape + (3,)  # for gain-mapped constants
        self.threshold_map_shape = self.map_shape + (2,)
        # constants
        self.gain_thresholds_gpu = cupy.empty(
            self.threshold_map_shape, dtype=np.float32
        )
        self.offset_map_gpu = cupy.zeros(self.gm_map_shape, dtype=np.float32)
        self.rel_gain_pc_map_gpu = cupy.ones(self.gm_map_shape, dtype=np.float32)
        # not gm_map_shape because it only applies to medium gain pixels
        self.md_additional_offset_gpu = cupy.zeros(self.map_shape, dtype=np.float32)
        self.rel_gain_xray_map_gpu = cupy.ones(self.map_shape, dtype=np.float32)
        self.badpixel_map_gpu = cupy.zeros(self.gm_map_shape, dtype=np.uint32)
        self.badpixel_mask_value = badpixel_mask_value  # TODO: make this configurable

        self.update_block_size((1, 1, 64))

    def _preview_preprocess_raw(self):
        return self.input_data_gpu[:, 0]

    def _preview_preprocess_corr(self):
        return self.processed_data_gpu

    def load_thresholds(self, threshold_map):
        # shape: y, x, memory cell, threshold 0 / threshold 1 / 3 gain values
        # TODO: do we need the gain values (in the constant) for anything?
        self.gain_thresholds_gpu.set(
            np.transpose(threshold_map[..., :2], (2, 1, 0, 3)).astype(np.float32)
        )

    def load_offset_map(self, offset_map):
        # shape: y, x, memory cell, gain stage
        self.offset_map_gpu.set(
            np.transpose(offset_map, (2, 1, 0, 3)).astype(np.float32)
        )

    def load_rel_gain_pc_map(self, slopes_pc_map):
        # pc has funny shape (11, 352, 128, 512) from file
        # this is (fi, memory cell, y, x)
        slopes_pc_map = slopes_pc_map.astype(np.float32)
        # the following may contain NaNs, though...
        hg_slope = slopes_pc_map[0]
        hg_intercept = slopes_pc_map[1]
        mg_slope = slopes_pc_map[3]
        mg_intercept = slopes_pc_map[4]
        # from agipdlib.py: replace NaN with median (per memory cell)
        # note: suffixes in agipdlib are "_m" and "_l", should probably be "_I"
        for naughty_array in (hg_slope, hg_intercept, mg_slope, mg_intercept):
            medians = np.nanmedian(naughty_array, axis=(1, 2))
            nan_bool = np.isnan(naughty_array)
            nan_cell, _, _ = np.where(nan_bool)
            naughty_array[nan_bool] = medians[nan_cell]

            # TODO: verify that this clamping should be done (as in agipdlib.py)
            too_low_bool = naughty_array < 0.8 * medians[:, np.newaxis, np.newaxis]
            too_low_cell, _, _ = np.where(too_low_bool)
            naughty_array[too_low_bool] = medians[too_low_cell]

            too_high_bool = naughty_array > 1.2 * medians[:, np.newaxis, np.newaxis]
            too_high_cell, _, _ = np.where(too_high_bool)
            naughty_array[too_high_bool] = medians[too_high_cell]

        frac_hg_mg = hg_slope / mg_slope
        md_additional_offset = (hg_intercept - mg_intercept * frac_hg_mg).astype(
            np.float32
        )
        rel_gain_map = np.ones(
            (3, self.constant_memory_cells, self.pixels_y, self.pixels_x),
            dtype=np.float32,
        )
        rel_gain_map[1] = rel_gain_map[0] * frac_hg_mg
        rel_gain_map[2] = rel_gain_map[1] * 4.48
        # TODO: enable overriding md_additional_offset_gpu based on user input
        self.rel_gain_pc_map_gpu.set(np.transpose(rel_gain_map, (1, 3, 2, 0)))
        self.md_additional_offset_gpu.set(np.transpose(md_additional_offset, (0, 2, 1)))

    def load_rel_gain_ff_map(self, slopes_ff_map):
        # constant shape: y, x, memory cell
        if slopes_ff_map.shape[2] == 2:
            # old format, is per pixel only (shape is y, x, 2)
            slopes_ff_map = np.broadcast_to(
                slopes_ff_map[..., 0][..., np.newaxis],
                (self.pixels_y, self.pixels_x, self.constant_memory_cells),
            )
        else:
            # TODO: maybe extend in case constant has too few memory cells
            # TODO: maybe divide by something because new constants are absolute
            ...
        # TODO: maybe clamp and replace NaNs like slopes_pc
        self.rel_gain_xray_map_gpu.set(np.transpose(slopes_ff_map).astype(np.float32))

    def load_badpixels_map(self, badpixels_map):
        print(f"Loading bad pixels with shape: {badpixels_map.shape}")
        # will simply OR with already loaded, does not take into account which ones
        # TODO: simple loading
        # TODO: allow configuring subset of bad pixels to care about
        # TODO: allow configuring value for masked pixels
        # TODO: inquire what "mask for double size pixels" means
        if len(badpixels_map.shape) == 3:
            # BadPixelsFF is not per gain stage...
            # TODO: broadcast?
            ...
        else:
            self.badpixel_map_gpu |= cupy.asarray(
                np.transpose(badpixels_map, (2, 1, 0, 3)), dtype=np.uint32
            )

    def flush_buffers(self):
        self.offset_map_gpu.fill(0)
        self.rel_gain_pc_map_gpu.fill(1)
        self.md_additional_offset_gpu.fill(0)
        self.rel_gain_xray_map_gpu.fill(1)
        self.badpixel_map_gpu.fill(0)

    # TODO: baseline shift

    def correct(self, flags):
        if flags & CorrectionFlags.BLSHIFT:
            raise NotImplementedError("Baseline shift not implemented yet")
        if self.gain_mode is not AgipdGainMode.ADAPTIVE_GAIN and (
            flags & CorrectionFlags.THRESHOLD
        ):
            raise ValueError("Cannot do gain thresholding in fixed gain mode")
        self.correction_kernel(
            self.full_grid,
            self.full_block,
            (
                self.input_data_gpu,
                self.cell_table_gpu,
                np.uint8(flags),
                self.default_gain,
                self.gain_thresholds_gpu,
                self.offset_map_gpu,
                self.rel_gain_pc_map_gpu,
                self.md_additional_offset_gpu,
                self.rel_gain_xray_map_gpu,
                self.badpixel_map_gpu,
                self.badpixel_mask_value,
                self.gain_map_gpu,
                self.processed_data_gpu,
            ),
        )

    def _init_kernels(self):
        kernel_source = self._kernel_template.render(
            {
                "pixels_x": self.pixels_x,
                "pixels_y": self.pixels_y,
                "data_memory_cells": self.memory_cells,
                "constant_memory_cells": self.constant_memory_cells,
                "input_data_dtype": utils.np_dtype_to_c_type(self.input_data_dtype),
                "output_data_dtype": utils.np_dtype_to_c_type(self.output_data_dtype),
                "corr_enum": utils.enum_to_c_template(CorrectionFlags),
            }
        )
        self.source_module = cupy.RawModule(code=kernel_source)
        self.correction_kernel = self.source_module.get_function("correct")


class BadPixelValues(enum.IntFlag):
    """The European XFEL Bad Pixel Encoding

    Straight from pycalibration's enum.py"""

    OFFSET_OUT_OF_THRESHOLD = 0b000000000000000000001  # bit 1
    NOISE_OUT_OF_THRESHOLD = 0b000000000000000000010  # bit 2
    OFFSET_NOISE_EVAL_ERROR = 0b000000000000000000100  # bit 3
    NO_DARK_DATA = 0b000000000000000001000  # bit 4
    CI_GAIN_OF_OF_THRESHOLD = 0b000000000000000010000  # bit 5
    CI_LINEAR_DEVIATION = 0b000000000000000100000  # bit 6
    CI_EVAL_ERROR = 0b000000000000001000000  # bit 7
    FF_GAIN_EVAL_ERROR = 0b000000000000010000000  # bit 8
    FF_GAIN_DEVIATION = 0b000000000000100000000  # bit 9
    FF_NO_ENTRIES = 0b000000000001000000000  # bit 10
    CI2_EVAL_ERROR = 0b000000000010000000000  # bit 11
    VALUE_IS_NAN = 0b000000000100000000000  # bit 12
    VALUE_OUT_OF_RANGE = 0b000000001000000000000  # bit 13
    GAIN_THRESHOLDING_ERROR = 0b000000010000000000000  # bit 14
    DATA_STD_IS_ZERO = 0b000000100000000000000  # bit 15
    ASIC_STD_BELOW_NOISE = 0b000001000000000000000  # bit 16
    INTERPOLATED = 0b000010000000000000000  # bit 17
    NOISY_ADC = 0b000100000000000000000  # bit 18
    OVERSCAN = 0b001000000000000000000  # bit 19
    NON_SENSITIVE = 0b010000000000000000000  # bit 20
    NON_LIN_RESPONSE_REGION = 0b100000000000000000000  # bit 21
