import pathlib

import cupy
import cupyx
import jinja2
import numpy as np

from . import utils


class BaseGpuRunner:
    """Class to handle GPU buffers and execution of CUDA kernels on image data

    All GPU buffers are kept within this class and it is intentionally very stateful.
    This generally means that you will want to load data into it and then do something.
    Typical usage in correct order:

    1. instantiate
    2. load constants
    3. load_data
    4. load_cell_table
    5. correct
    6a. reshape (only here does data transfer back to host)
    6b. compute_preview (optional)

    repeat from 2. or 3.

    In case no constants are available / correction is not desired, can skip 3 and 4 and
    pass CorrectionFlags.NONE to correct(...). Generally, user must handle which
    correction steps are appropriate given the constants loaded so far.
    """

    def __init__(
        self,
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells,
        output_transpose=(2, 1, 0),  # default: memorycells-fast
        input_data_dtype=np.uint16,
        output_data_dtype=np.float32,
    ):
        _src_dir = pathlib.Path(__file__).absolute().parent
        # subclass must define _kernel_source_filename
        with (_src_dir / self._kernel_source_filename).open("r") as fd:
            self._kernel_template = jinja2.Template(fd.read())

        self.pixels_x = pixels_x
        self.pixels_y = pixels_y
        self.memory_cells = memory_cells
        self.output_transpose = output_transpose
        if constant_memory_cells == 0:
            # if not set, guess same as input; may save one recompilation
            self.constant_memory_cells = memory_cells
        else:
            self.constant_memory_cells = constant_memory_cells
        self.output_shape = utils.shape_after_transpose(
            self.processed_shape, self.output_transpose
        )
        # preview will only be single memory cell
        self.preview_shape = (self.pixels_x, self.pixels_y)
        self.input_data_dtype = input_data_dtype
        self.output_data_dtype = output_data_dtype

        self._init_kernels()

        # reuse buffers for input / output
        self.cell_table_gpu = cupy.empty(self.memory_cells, dtype=np.uint16)
        self.input_data_gpu = cupy.empty(self.input_shape, dtype=input_data_dtype)
        self.processed_data_gpu = cupy.empty(
            self.processed_shape, dtype=output_data_dtype
        )
        self.reshaped_data_gpu = cupy.empty(self.output_shape, dtype=output_data_dtype)
        self.preview_raw = cupyx.empty_pinned(self.preview_shape, dtype=np.float32)
        self.preview_corrected = cupyx.empty_pinned(
            self.preview_shape, dtype=np.float32
        )

    # functions to get data from respective buffers to cell, x, y shape for preview computation
    # TODO: handle shape juggling programmatically, removing need for these two helpers
    def _preview_preprocess_raw(self):
        """Should return view of self.input_data_gpu with shape (cell, x, y)"""
        raise NotImplementedError()

    def _preview_preprocess_corr(self):
        """Should return view of self.processed_data_gpu with shape (cell, x, y)"""
        raise NotImplementedError()

    def flush_buffers(self):
        """Optional reset GPU buffers (implement in appropriate subclasses"""
        pass

    def correct(self, flags):
        """Correct (already loaded) image data according to flags

        Subclass must define this method. It should assume that image data, cell table,
        and other data (including constants) has already been loaded. It should
        probably run some GPU kernel and output should go into self.processed_data_gpu.

        Keep in mind that user only gets output from compute_preview or reshape
        (either of these should come after correct).

        The submodules providing subclasses should have some IntFlag enums defining
        which flags are available to pass along to the kernel. A zero flag should allow
        the kernel to do no actual correction - but still copy the data between buffers
        and cast it to desired output type.

        """
        raise NotImplementedError()

    def reshape(self, out=None):
        """Move axes to desired output order

        The out parameter is passed directly to the get function of GPU array: if
        None, then a new ndarray (in host memory) is returned. If not None, then data
        will be loaded into the provided array, which must match shape / dtype.
        """
        # TODO: avoid copy
        if self.output_transpose is None:
            self.reshaped_data_gpu = cupy.ascontiguousarray(
                cupy.squeeze(self.processed_data_gpu)
            )
        else:
            self.reshaped_data_gpu = cupy.ascontiguousarray(
                cupy.transpose(
                    cupy.squeeze(self.processed_data_gpu), self.output_transpose
                )
            )
        return self.reshaped_data_gpu.get(out=out)

    def load_data(self, raw_data):
        self.input_data_gpu.set(np.squeeze(raw_data))

    def load_cell_table(self, cell_table):
        self.cell_table_gpu.set(cell_table)

    def compute_preview(self, preview_index):
        """Generate single slice or reduction preview of raw and corrected data

        Special values of preview_index are -1 for max (select max integrated intensity
        frame), -2 for mean, -3 for sum, and -4 for stdev (across cells).

        Note that preview_index is taken from data without checking cell table.
        Caller has to figure out which index along memory cell dimension they
        actually want to preview in case it needs to be a specific pulse.

        Will reuse data from corrected output buffer. Therefore, correct(...) must have
        been called with the appropriate flags before compute_preview(...).
        """

        if preview_index < -4:
            raise ValueError(f"No statistic with code {preview_index} defined")
        elif preview_index >= self.memory_cells:
            raise ValueError(f"Memory cell index {preview_index} out of range")

        # TODO: enum around reduction type
        for (preprocces, output_buffer) in (
            (self._preview_preprocess_raw, self.preview_raw),
            (self._preview_preprocess_corr, self.preview_corrected),
        ):
            image_data = preprocces()
            if preview_index >= 0:
                # TODO: change axis order when moving reshape to after correction
                image_data[preview_index].astype(np.float32).get(out=output_buffer)
            elif preview_index == -1:
                # TODO: select argmax independently for raw and corrected?
                # TODO: send frame sums somewhere to compute global max frame
                max_index = cupy.argmax(
                    cupy.sum(image_data, axis=(1, 2), dtype=cupy.float32)
                )
                image_data[max_index].astype(np.float32).get(out=output_buffer)
            elif preview_index in (-2, -3, -4):
                stat_fun = {-2: cupy.mean, -3: cupy.sum, -4: cupy.std}[preview_index]
                stat_fun(image_data, axis=0, dtype=cupy.float32).get(out=output_buffer)
        return self.preview_raw, self.preview_corrected

    def update_block_size(self, full_block, target_shape=None):
        """Compute grid such that thread block grid covers target shape

        Execution is scheduled with 3d "blocks" of CUDA threads, tuning can affect
        performance. Correction kernels are "monolithic" for simplicity (i.e. each
        logical thread handles one entry in output data), so in each dimension we
        parallelize, grid * block >= length.

        Note that individual kernels must themselves check whether they go out of
        bounds; grid dimensions get rounded up in case ndarray size is not multiple of
        block size.

        """
        if target_shape is None:
            target_shape = self.processed_shape
        assert len(full_block) == 3
        self.full_block = tuple(full_block)
        self.full_grid = tuple(
            utils.ceil_div(a_length, block_length)
            for (a_length, block_length) in zip(target_shape, full_block)
        )
