import re

import pycuda.driver
import pycuda.gpuarray

_gpuptr_re = re.compile(
    r"GPUPTR:(?P<gpu_pointer>\w+)" r"DEVID:(?P<device_id>.+)" r"SHAPE:(?P<shape>.+)"
)


def get_shape_from_ipc_handle(handle_string):
    match = _gpuptr_re.match(handle_string)
    return tuple(int(num) for num in match.group("shape").split(","))


class IPCGPUArray:
    """Context manager providing a GPUArray opened from string encoding IPC handle

    Arguments:
    handle_string: String encoding a "GPU pointer" (IPC address) plus some more
    stuff.  This is "parsed" using _gpuptr_re.
    dtype: self-explanatory (but make sure it is correct)
    aray shape is parsed from the handle_string
    """

    def __init__(self, handle_string, dtype, gpu_pointer_re=None):
        match = _gpuptr_re.match(handle_string)
        assert match is not None

        self.dtype = dtype
        self.handle_address = bytearray.fromhex(match.group("gpu_pointer"))
        self.shape = tuple(int(num) for num in match.group("shape").split(","))
        # assuming contiguous C-order strides probably
        # TODO: smarter

        self.open_handle = None
        self.gpu_array = None

    def __enter__(self):
        self.open_handle = pycuda.driver.IPCMemoryHandle(self.handle_address)
        self.gpu_array = pycuda.gpuarray.GPUArray(
            self.shape, dtype=self.dtype, gpudata=self.open_handle
        )
        return self.gpu_array

    def __exit__(self, t, v, tb):
        self.open_handle.close()


class GPUContextContext:
    def __init__(self, gpu_context):
        self.gpu_context = gpu_context

    def __enter__(self):
        self.gpu_context.push()
        return self.gpu_context

    def __exit__(self, t, v, tb):
        self.gpu_context.pop()
